#include "task_manager.h"
#include "task.h"

#include <queue>

using std::queue;

TaskManager::TaskManager()
{
    fp_task = stdin;
    fp_dependency = stdin;
    fp_processor = stdin;
    fp_schedule = stdin;

    adjacency_table = NULL;
    processor = NULL;

}

TaskManager::~TaskManager()
{
    for (int i = 0; i < task_num; i++)
	delete[] adjacency_table[i];
    delete[] adjacency_table;

    delete[] processor;
}

void TaskManager::make_graph(FILE *fp_task, FILE *fp_dependency, 
			     FILE *fp_processor, FILE *fp_schedule)
{
    this->fp_task = fp_task;
    this->fp_dependency = fp_dependency;
    this->fp_processor = fp_processor;
    this->fp_schedule = fp_schedule;

    __configure();
}

void TaskManager::print_output(FILE *fp_output)
{
    fprintf(fp_output, "Connected table:\n");
    for (int i = 0; i < task_num; i++)
    {
	fprintf(fp_output, "\t\t{");
	for (int j = 0; j < task_num - 1; j++)
	    fprintf(fp_output, "%d, ", adjacency_table[i][j]);
	fprintf(fp_output, "%d},\n", adjacency_table[i][task_num - 1]);
    }

    // Print time_cost table.
    fprintf(fp_output, "Time Cost:\n");
    for (int i = 0; i < task_num; i++)
    {
	int duration = tasks[i].get_duration();
	fprintf(fp_output, "\t\t{");
	for (int j = 0; j < processor_num; j++)
	    fprintf(fp_output, "%d, ", duration);
	fprintf(fp_output, "\b\b},\n");
    }

    // Print assigned table.
    fprintf(fp_output, "Assigned Table:\n");
    fprintf(fp_output, "\t{");
    for (int i = 0; i < task_num; i++)
    {
	fprintf(fp_output, "%d, ", assign_table[i]);
    }
    fprintf(fp_output, "\b\b}\n");

    // Print the tasks which is delayed caused by the power limination.
    int delayed_task_num = delayed_task.size();
    fprintf(fp_output, "Delayed Tasks:\n");
    for (int i = 0; i < delayed_task_num; i++)
    {
	fprintf(fp_output, "%d ", delayed_task[i]);
    }
    fprintf(fp_output, "\n");

    // Print the power and variation settings.
    fprintf(fp_output,"\t\t{");
    for (int i = 0; i < processor_type; i++)
    {
	for (int j = 0; j < processor[i]; j++)
	{
	    fprintf(fp_output,"\t\t{}");
	}
    }
}

void TaskManager::__configure()
{
    // Read the task file.
    Task task;
    int id;
    int type;
    int duration;

    while (fscanf(fp_task, "%d%d%d", &id, &type, &duration) > 0)
    {
	task.set_id(id);
	task.set_type(type - 1);
	task.set_duration(duration);
	tasks.push_back(task);
    }
    task_num = tasks.size();
	
    // Read the dependency file.
    int source;
    int destination;

    adjacency_table = new int *[task_num];
    for (int i = 0; i < task_num; i++)
    {
	adjacency_table[i] = new int[task_num];
    }

    for (int i = 0; i < task_num; i++)
    {
	for (int j = 0; j < task_num; j++)
	    adjacency_table[i][j] = 0;
    }

    while (fscanf(fp_dependency, "%d%d", &source, &destination) > 0)
    {
	adjacency_table[source - 1][destination - 1] = 1;
    }

    // Read the processor description file.
    fscanf(fp_processor, "%d%d", &processor_num, &processor_type);
    processor = new int[processor_type];
    for (int i = 0; i < processor_type; i++)
	fscanf(fp_processor, "%d", &processor[i]);

    // Read the schedule file.
    int start_time;
    int processor_id;
    while (fscanf(fp_schedule, "%d%d%d", &id, &start_time, &processor_id) > 0)
    {
	Task *p = &tasks[id - 1];
	p->set_real_start_time(start_time);
	assign_table.push_back(__get_processor_id(p->get_type(), processor_id));
    }

    __check_delay();
}

int TaskManager::__get_processor_id(int type, int id) const
{
    int processor_id = 0;
    for (int i = 0; i < type; i++)
	processor_id += processor[i];
    processor_id += id;
	
    return processor_id;
}

void TaskManager::__check_delay()
{	
    for (int col = 0; col < task_num; col++)
    {
	for (int row = 0; row < task_num; row++)
	{
	    if (adjacency_table[row][col] == 1)
	    {
		int end_time = tasks[row].get_real_start_time()
		    + tasks[row].get_duration();
		tasks[row].set_end_time(end_time);
		if (tasks[col].get_earliest_start_time() < end_time)
		    tasks[col].set_earliest_start_time(end_time);
	    }
	}
    }
	
    // Set end time.
    for (int i = 0; i < task_num; i++)
    {
	tasks[i].set_end_time(tasks[i].get_real_start_time() +
			      tasks[i].get_duration());
    }


    for (int i = 0; i < task_num; i++)
    {
	printf("ID:%d | Earliest:%d / Real:%d / End:%d\n", 
	       tasks[i].get_id(),
	       tasks[i].get_earliest_start_time(), 
	       tasks[i].get_real_start_time(),
	       tasks[i].get_end_time());


	if (tasks[i].get_earliest_start_time() < tasks[i].get_real_start_time())
	    delayed_task.push_back(tasks[i].get_id());
    }

    printf("\n----------\n");
    for (vector<int>::iterator it = delayed_task.begin();
	 it != delayed_task.end(); it++)
    {
	for (int j = 0; j < task_num; j++)
	{
	    // If task[i] starts after task[j] ends, add an dependency.
	    if (tasks[j].get_end_time() <= tasks[*it - 1].get_real_start_time())
	    {
		adjacency_table[j][*it - 1] = 1;
	    }
	}
    }
}
